// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import java.util.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import hep.wired.util.SwingEventMulticaster;

/**
 * Adapter for implementing {@link ConditionDefinition}.
 *
 * @author Dmitry Onoprienko
 * @version $Id: AbstractConditionDefinition.java 13824 2011-05-04 18:30:02Z onoprien $
 */
abstract public class AbstractConditionDefinition implements ConditionDefinition {

// -- Private parts : ----------------------------------------------------------

  protected String _name;
  protected String _unit;
  protected String _description;

  protected boolean _valueIsAdjusting;
  private ChangeListener _listener;
  private final ChangeEvent _defaultEvent = new ChangeEvent(this);
  protected HashSet<Condition> _conditions;

// -- Construction and initialization : ----------------------------------------

  public AbstractConditionDefinition(String name) {
    this(name, null, null);
  }

  public AbstractConditionDefinition(String name, String unit, String description) {
    _name = name;
    _unit = unit;
    _description = description;
    _conditions = new HashSet();
  }

// -- Name and description : ---------------------------------------------------

  /** Returns the name of this definition. */
  public String getName() {
    return _name;
  }

  /** Returns the unit name for conditions based on numerical value. */
  public String getUnit() {
    return _unit;
  }

  /** Returns a brief description of the condition. */
  public String getDescription() {
    return _description;
  }


// -- Creating Conditions : ----------------------------------------------------

  /**
   * Returns a new {@link Condition} object linked to this ConditionDefinition.
   * The newly created Condition is registered with this ConditionDefinition.
   * Conditions created with this method are not enabled and not inverted.
   */
  public Condition createCondition() {
    return createCondition(false, false);
  }

  /** Unregisters the specified Condition. */
  public void removeCondition(Condition condition) {
    _conditions.remove(condition);
  }

  /** 
   * Called when this definition is being deleted.
   * Tells Cuts of all registered conditions that they need to delete themselves.
   */
  public void remove() {
    for (Condition c : new ArrayList<Condition>(_conditions)) {
      Cut cut = c.getCut();
      if (cut != null) cut.remove();
    }
  }


// -- Handling listeners : -----------------------------------------------------

  /** Adds change listener. */
  public void addChangeListener(ChangeListener listener) {
    listener = SwingEventMulticaster.add(_listener, listener);
  }

  /** Removes change listener. */
  public void removeChangeListener(ChangeListener listener) {
    listener = SwingEventMulticaster.remove(_listener, listener);
  }

  /**
   * Sets <tt>valueIsAdjusting</tt> property.
   * Change events are not generated while the property is set to <tt>false</tt>.
   * An event is generated when the property is set to <tt>true</tt>.
   */
  public void setValueIsAdjusting(boolean isValueAdjusting) {
    _valueIsAdjusting = isValueAdjusting;
    fireStateChanged();
  }

  /**
   * Returns <tt>valueIsAdjusting</tt> property.
   * Change events are not generated while the property is set to <tt>false</tt>.
   */
  public boolean getValueIsAdjusting() {
    return _valueIsAdjusting;
  }

  protected void fireStateChanged() {
    if (!_valueIsAdjusting) {
//      List<Condition> conditions = new ArrayList(_conditions.keySet());
//      for (Condition c : conditions) c.stateChanged(_defaultEvent);
      if (_listener != null) _listener.stateChanged(_defaultEvent);
    }
  }

  protected void fireStateChanged(ChangeEvent event) {
    if (!_valueIsAdjusting) {
//      List<Condition> conditions = new ArrayList(_conditions.keySet());
//      for (Condition c : conditions) c.stateChanged(event);
      if (_listener != null) _listener.stateChanged(event);
    }
  }

}
