// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import javax.swing.event.ChangeListener;

import org.freehep.util.Value;

/**
 * Definition of a condition used to filter displayable objects based on a single attribute value.
 * 
 * @author Mark Donszelmann
 * @author Dmitry Onoprienko
 * @version $Id: ConditionDefinition.java 13824 2011-05-04 18:30:02Z onoprien $
 */

public interface ConditionDefinition {

  enum Type {BOOLEAN, INTEGER, LONG, DOUBLE, STRING}

// -- Name and description : ---------------------------------------------------

  /** Returns the name of this definition. */
  String getName();

  /** Returns the unit name for conditions based on numerical value. */
  String getUnit();

  /** Returns a brief description of the condition. */
  String getDescription();

  /** Returns the type of this condition. */
  Type getType();


// -- Creating/Removing Conditions : -------------------------------------------

  /**
   * Returns a new {@link Condition} object linked to this ConditionDefinition.
   * The newly created Condition is registered with this ConditionDefinition.
   * Conditions created with this method are not enabled and not inverted.
   */
  Condition createCondition();

  /**
   * Returns a new {@link Condition} object linked to this ConditionDefinition.
   * The newly created Condition is registered with this ConditionDefinition.
   */
  Condition createCondition(boolean enabled, boolean inverted);

  /** Unregisters the specified Condition. */
  void removeCondition(Condition condition);

  /** Signals to all registered conditions that this definition is being deleted.*/
  void remove();


// -- Domain : -----------------------------------------------------------------

  /**
   * Sets the domain of possible values for the condition.
   * The actual type of object used to describe domain and its interpretation depends on
   * the type of the condition. Calling this method with <tt>null</tt> argument should reset
   * the domain to an empty set unless the domain is fixed (like in boolean conditions).
   */
  void setDomain(Object domain);

  /**
   * Returns the domain of possible values for the condition.
   * The actual type of object used to describe domain and its interpretation depends on
   * the type of the condition.
   */
  Object getDomain();

  /** Updates the domain to ensure it contains the specified value. */
  void addToDomain(Value value);


// -- Handling listeners : -----------------------------------------------------

  /** Adds change listener. */
  void addChangeListener(ChangeListener listener);

  /** Removes change listener. */
  void removeChangeListener(ChangeListener listener);

  /**
   * Sets <tt>valueIsAdjusting</tt> property.
   * Change events are not generated while the property is set to <tt>false</tt>.
   * An event is generated when the property is set to <tt>true</tt>.
   */
  void setValueIsAdjusting(boolean isValueAdjusting);

  /**
   * Returns <tt>valueIsAdjusting</tt> property.
   * Change events are not generated while the property is set to <tt>false</tt>.
   */
  boolean getValueIsAdjusting();

}
