// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import hep.wired.util.SwingEventMulticaster;
import java.util.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.freehep.util.Value;

/**
 * Cut that can be applied to displayable objects.
 *
 * @author Dmitry Onoprienko
 * @version $Id: Cut.java 13832 2011-05-10 23:25:08Z onoprien $
 */
public class Cut {
  
  public enum Result {PASS, FAIL, IGNORE}

// -- Private parts : ----------------------------------------------------------

  private String _name;
  private List<Condition> _conditions;
  private ArrayList<CutSet> _sets;

  static protected final String _defaultName = "New cut";
  protected final ChangeEvent _defaultEvent = new ChangeEvent(this);
  private ChangeListener _listener;
  protected boolean _valueIsAdjusting;
  protected boolean _eventsPending;

// -- Construction : -----------------------------------------------------------

  public Cut() {
    this(_defaultName);
  }

  public Cut(String name) {
    _name = name;
    _conditions = new ArrayList();
    _sets = new ArrayList();
    _valueIsAdjusting = false;
    _eventsPending = false;
  }

// -- Getters : ----------------------------------------------------------------

  /** Returns the name of the cut. */
  public String getName() {
    return _name;
  }

  /** Returns condition specified by the index. */
  public Condition getCondition(int index) {
    return _conditions.get(index);
  }

  /** Returns the list of conditions in this cut. */
  public List<Condition> getConditions() {
    return Collections.unmodifiableList(_conditions);
  }

  /** Returns the number of conditions in this cut. */
  public int size() {
    return _conditions.size();
  }

  /** Returns a set of enabled conditions. */
  public Set getEnabledConditions() {
    Set<Condition> conditionSet = new HashSet();
    for (Condition condition : _conditions) {
      if (condition.isEnabled()) {
        conditionSet.add(condition);
      }
    }
    return conditionSet;
  }

// -- Setters : ----------------------------------------------------------------

  /** Sets the name of this cut. */
  public void setName(String newName) {
    if (_name.equals(newName)) return;
    _name = newName;
    fireStateChanged(new ChangeEvent(this));
  }

  /**
   * Adds the specified CutSet to the list of known sets.
   * CutSet will be notified of any changes in this Cut.
   */
  public void addCutSet(CutSet set) {
    _sets.add(set);
  }

  /**
   * Removes the specified CutSet from the list of known sets.
   * CutSet will be notified of any changes in this Cut.
   */
  public void removeCutSet(CutSet set) {
    _sets.remove(set);
  }

  /** Adds condition to this cut. */
  public void add(Condition condition) {
    _conditions.add(condition);
    condition.setCut(this);
    fireStateChanged(_defaultEvent);
  }

  public void addAll(Collection<? extends Condition> conditions) {
    _conditions.addAll(conditions);
    for (Condition c : conditions) c.setCut(this);
    fireStateChanged(_defaultEvent);
  }

  /** Removes the specified condition from this cut. */
  public void remove(Condition condition) {
    if (condition != null) {
      int index = _conditions.indexOf(condition);
      if (index >= 0) {
        _conditions.remove(condition);
        condition.setCut(null);
        condition.getDefinition().removeCondition(condition);
        fireStateChanged(_defaultEvent);
      }
    }
  }

  /** Removes all conditions from this cut. */
  public void removeAll() {
    for (Condition c : _conditions) {
      c.setCut(null);
      c.getDefinition().removeCondition(c);
    }
    _conditions.clear();
    fireStateChanged(_defaultEvent);
  }

  /** Removes disabled conditions from this cut. */
  public void removeDisabled() {
    HashSet<Condition> dc = new HashSet();
    for (Condition c : _conditions) {
      if (!c.isEnabled()) {
        c.setCut(null);
        c.getDefinition().removeCondition(c);
        dc.add(c);
      }
    }
    _conditions.removeAll(dc);
    fireStateChanged(_defaultEvent);
  }

  /**
   * Removes conditions whose definitions are not in the supplied collection, creates and adds disabled condition
   * for every definition from the collection that is not already used by this cut.
   */
  public void synchronize(Collection<ConditionDefinition> definitions) {
    Set<ConditionDefinition> currentDefs = new HashSet();
    for (Condition c : _conditions) currentDefs.add(c.getDefinition());
    HashSet<ConditionDefinition> toAdd = new HashSet(definitions);
    toAdd.removeAll(currentDefs);
    currentDefs.removeAll(definitions);
    HashSet<Condition> dc = new HashSet();
    for (Condition c : _conditions) {
      if (currentDefs.contains(c.getDefinition())) {
        c.setCut(null);
        c.getDefinition().removeCondition(c);
        dc.add(c);
      }
    }
    _conditions.removeAll(dc);
    for (ConditionDefinition def : toAdd) {
      Condition c = def.createCondition();
      c.setCut(this);
      _conditions.add(c);
    }
    fireStateChanged(_defaultEvent);
  }

  /**  Resets all conditions to their default state. */
  public void reset() {
    setValueIsAdjusting(true);
    for (Condition condition : _conditions) {
      condition.reset();
    }
    setValueIsAdjusting(false);
  }

  /**
   * Called when this Cut needs to be deleted.
   * Removes itself from all {@link CutSet}s, unregisters its {@link Condition}s from {@link ConditionDefinition}s.
   */
  public void remove() {
    for (CutSet set : _sets) set.remove(this);
    for (Condition c : _conditions) c.getDefinition().removeCondition(c);
  }

// -- Apply cut : --------------------------------------------------------------

  /**
   * Tells the client whether this cut is satisfied by the supplied <tt>data</tt>.
   * If the cut is not inverted, returns PASS if the data satisfies all applicable conditions, FAIL if
   * the data fails at least one condition, IGNORE is there are no enabled applicable conditions.
   * If the cut is inverted, returns PASS if the data fails at least one condition, FAIL if
   * all applicable conditions are satisfied, IGNORE is there are no enabled applicable conditions.
   *
   * The condition is deemed applicable if <tt>data</tt> maps its name to non-null <tt>Value</tt>
   * with compatible type.
   *
   * @param data  Map of attribute names to values.
   * @param invert  True if this cut should be inverted.
   * @return  PASS if the data satisfies this cut, FAIL if the data fails the cut, and IGNORE if the cut is inapplicable.
   */
  public Result pass(Map<String,Value> data, boolean invert) {
    if (data == null) return Result.IGNORE;
    Result out = Result.IGNORE;
    for (Condition condition : _conditions) {
      if (!condition.isEnabled()) continue;
      Value value = data.get(condition.getDefinition().getName());
      if ((value != null) && (value.getType() != null) & (value.getType() != Object.class)) {
        if (invert) {
          if (condition.pass(value)) {
            return Result.FAIL;
          } else {
            out = Result.PASS;
          }
        } else {
          if (!condition.pass(value)) {
            return Result.FAIL;
          } else {
            out = Result.PASS;
          }
        }
      }
    }
    return out;
  }


  /**
   * Tells the client whether this cut is satisfied by the supplied data.
   * The implementation provided by this class forwards the call to
   * {@link #pass(java.util.Map, boolean)}, ignoring <tt>moreData</tt> parameter.
   * Subclasses may override to implement more complex filtering.
   *
   * @param data  Map of attribute names to values.
   * @param moreData  Additional data to be used in filtering.
   * @param invert  True if this cut should be inverted.
   * @return  PASS if the data satisfies this cut, FAIL if the data fails the cut, and IGNORE if the cut is inapplicable.
   */
  public Result pass(Map<String,Value> data, Object moreData, boolean invert) {
    return pass(data, invert);
  }
  
// -- Handling events and listeners : ------------------------------------------

  public void addChangeListener(ChangeListener a) {
    _listener = SwingEventMulticaster.add(_listener, a);
  }

  public void removeChangeListener(ChangeListener r) {
    _listener = SwingEventMulticaster.remove(_listener, r);
  }

  public void fireStateChanged(ChangeEvent e) {
    if (_valueIsAdjusting) {
      _eventsPending = true;
    } else {
      for (CutSet set : _sets) {
        set.stateChanged(e);
      }
      if (_listener != null) _listener.stateChanged(e);
    }
  }

  public void fireStateChanged() {
    fireStateChanged(_defaultEvent);
  }

  /**
   * Sets <tt>valueIsAdjusting</tt> property.
   * If <tt>valueIsAdjusting</tt> is <tt>false</tt>, any calls to <tt>fireStateChanged()</tt> methods
   * will result in listeners being notified immediately. If <tt>valueIsAdjusting</tt> is <tt>true</tt>,
   * notification will be postponed until the next call to <tt>setValueIsAdjusting(false)</tt>, and
   * a single event with this cut as a source will be sent then.
   */
  public void setValueIsAdjusting(boolean valueIsAdjusting) {
    if (!valueIsAdjusting && _valueIsAdjusting && _eventsPending) {
      _eventsPending = false;
      fireStateChanged(_defaultEvent);
    }
    _valueIsAdjusting = valueIsAdjusting;
  }

// -- Overriding Object : ------------------------------------------------------

  public String toString() {
    return "Cut " + _name + " " + hashCode();
  }
}
