package hep.wired.cut;

import java.text.DecimalFormat;

/**
 *
 * @author Dmitry Onoprienko
 * @version $Id: $
 */
public final class DoubleSliderEditor extends SliderEditor {

// -- Private parts : ----------------------------------------------------------

  double _min, _max;  // Domain bounds as accessible by the slider (logs of bounds if log scale)

  private double _scale;
  boolean _log;

  private DecimalFormat _formatExp = new DecimalFormat("0.0#E0");
  private DecimalFormat _formatPlain = new DecimalFormat("###0.0###");

// -- Construction and initialization : ----------------------------------------

  public DoubleSliderEditor(SliderPanel editor) {
    super(editor);
  }

// -- Setters : ----------------------------------------------------------------

  void setDomain() {
    double[] domain = (double[]) getCondition().getDefinition().getDomain();
    _min = Math.min(domain[0], getCondition().getMinimum());
    _max = Math.max(domain[1], getCondition().getMaximum());
    if (_min <= Double.MIN_NORMAL) {
      _log = false;
      _panel.getMarkBox().setEnabled(false);
    } else {
      _panel.getMarkBox().setEnabled(true);
    }
    if (_log) {
      _min = Math.floor(Math.log10(_min));
      _max = Math.ceil(Math.log10(_max));
      if (_max - _min < .1) {
        _min -=1.;
        _max +=1.;
      }
    } else {
      double d = (_max - _min) * .02;
      if (d <= Double.MIN_NORMAL) {
        _min = Math.floor(_min);
        _max = Math.ceil(_max);
        if (_max - _min < .1) {
          _min -= 1;
          _max += 1.;
        }
      } else {
        _min = (_min > 0. && _min - d < 0) ? 0. : _min - d;
        _max += d;
      }
    }
  }

  public void setScale() {
    _scale = (_max - _min) / _slider._size;
  }


// -- Updates from Slider : ----------------------------------------------------

  /** Called by ConditionSlider when its knobs have been moved. */
  public void updateFromSlider(int low, int high, boolean valueIsAdjusting) {
    double min = convertFromSlider(low);
    _panel.getMarkBox().setEnabled(min > Double.MIN_NORMAL);
    double max = convertFromSlider(high);
    if (valueIsAdjusting == _applyImmediately) getCondition().setRange(min, max);
    if (valueIsAdjusting) {
      _slider.updateMarkPositions(low, high);
      _panel.getLowField().setText(convertToField(min));
      _panel.getHighField().setText(convertToField(max));
    }
  }

  /** Called by ConditionSlider when its low knob has been moved. */
  public void updateFromSliderLow(int low, boolean valueIsAdjusting) {
    double min = convertFromSlider(low);
    _panel.getMarkBox().setEnabled(min > Double.MIN_NORMAL);
    if (valueIsAdjusting == _applyImmediately) getCondition().setMinimum(min);
    if (valueIsAdjusting) {
      _slider.updateMarkPositionLow(low);
      _panel.getLowField().setText(convertToField(min));
    }
  }

  /** Called by ConditionSlider when its high knob has been moved. */
  public void updateFromSliderHigh(int high, boolean valueIsAdjusting) {
    double max = convertFromSlider(high);
    if (valueIsAdjusting == _applyImmediately) getCondition().setMaximum(max);
    if (valueIsAdjusting) {
      _slider.updateMarkPositionHigh(high);
      _panel.getHighField().setText(convertToField(max));
    }
  }


// -- Updates from panel : -----------------------------------------------------

  void setMark(boolean state) {
    _log = state;
    setDomain();
    _slider.doLayout();
  }

  void reset() {
    getCondition().reset();
    setDomain();
    _slider.doLayout();
    updateFields();
  }

  void updateFromLowField(String low) throws NumberFormatException {
    double min = Double.parseDouble(low);
    boolean _maxChanged = false;
    if (min > getCondition().getMaximum()) {
      getCondition().setRange(min, min);
      updateHighField();
      _maxChanged = true;
    } else {
      getCondition().setMinimum(min);
    }
    if (min < (_log ? Math.pow(10., _min) : _min) || min > (_log ? Math.pow(10., _max) : _max)) {
      setDomain();
      _slider.doLayout();
    } else {
      _panel.getMarkBox().setEnabled(min > Double.MIN_NORMAL);
      if (_maxChanged) {
        updateSlider();
      } else {
        updateSliderLow();
      }
    }
  }

  void updateFromHighField(String high) throws NumberFormatException {
    double max = Double.parseDouble(high);
    boolean _minChanged = false;
    if (max < getCondition().getMinimum()) {
      getCondition().setRange(max, max);
      updateLowField();
      _minChanged = true;
    } else {
      getCondition().setMaximum(max);
    }
    if (max > (_log ? Math.pow(10., _max) : _max) || max < (_log ? Math.pow(10., _min) : _min)) {
      setDomain();
      _slider.doLayout();
    } else {
      if (_minChanged) {
        _panel.getMarkBox().setEnabled(max > Double.MIN_NORMAL);
        updateSlider();
      } else {
        updateSliderHigh();
      }
    }
  }

// -- Getters : ----------------------------------------------------------------

  public DoubleCondition getCondition() {
    return (DoubleCondition) _condition;
  }

  boolean getMark() {
    return _log;
  }

  boolean isLogScalePossible() {
    double[] domain = (double[]) getCondition().getDefinition().getDomain();
    return Math.min(getCondition().getMinimum(), domain[0]) > Double.MIN_NORMAL;
  }

  /** Returns low bound of the range in slider coordinates. */
  public int getLow() {
    return convertToSlider(getCondition().getMinimum());
  }

  /** Returns high bound of the range in slider coordinates. */
  public int getHigh() {
    return convertToSlider(getCondition().getMaximum());
  }

  /** Returns String representation of low bound. */
  String getLowString() {
    return convertToField(getCondition().getMinimum());
  }

  /** Returns String representation of high bound. */
  String getHighString() {
    return convertToField(getCondition().getMaximum());
  }

  
// -- Utility functions : ------------------------------------------------------

  int convertToSlider(double x) {
    return (int) (_log ? Math.round((_max - Math.log10(x)) / _scale) : Math.round((_max - x) / _scale));
  }

  double convertFromSlider(int x) {
    return _log ? Math.pow(10., (_max - x * _scale)) : _max - x * _scale;
  }

  String convertToField(double x) {
    double ax = Math.abs(x);
    if ((ax < 10000. && ax >= 0.01) || ax < Double.MIN_NORMAL) {
      String out = _formatPlain.format(x);
      int n = out.length();
      int significantDigits = 0;
      boolean afterPoint = false;
      int fromPoint = 0;
      int i = -1;
      while (++i < n && (fromPoint < 1 || significantDigits < 4)) {
        char c = out.charAt(i);
        if (c == '.') {
          afterPoint = true;
        } else {
          if (afterPoint) fromPoint++;
          if (c != '0' && c != '-') significantDigits++;
        }
      }
      return out.substring(0, i);
    } else {
      return _formatExp.format(x).toLowerCase();
    }
  }

// -- Creating ruller for Slider : ---------------------------------------------

  Slider.Ruler createRuler() {
    return _log ? createLogRuler() : createLinearRuler();
  }

  private Slider.Ruler createLinearRuler() {

    Slider.Ruler ruller = _slider.getNewRuler();

    double range = (_max - _min);
    double minDelta = range/ruller.nDiv;
    double order = Math.floor(Math.log10(range));
    double delta = Math.pow(10, order);
    double major = delta;
    double minor = 0;
    double digit = range/delta;
    if (digit < 2) {
      order--;
      major = Math.pow(10, order);
    } else if (digit < 6) {
      major = delta/5.;
    } else {
      major = delta;
    }
    delta = major;
    while (delta < minDelta) delta += major;
    minor = major/5.;

    double n = Math.floor(_min/delta);
    for (double x = delta*n; x < _max; x += delta) {
      int pos = convertToSlider(x);
      if (pos < _slider._size + 2) {
        if (Math.abs(x/delta) < 0.001) x = 0.;
        ruller.addLabel(pos, convertToField(x));
      }
    }

    n = Math.floor(_min/major);
    for (double x = major*n; x < _max; x += major) {
      int pos = convertToSlider(x);
      if (pos < _slider._size + 2) ruller.addMajorTick(pos);
    }

    n = Math.floor(_min/minor);
    for (double x = minor*n; x < _max; x += minor) {
      int pos = convertToSlider(x);
      if (pos < _slider._size + 2) ruller.addMinorTick(pos);
    }

    return ruller;
  }

  private Slider.Ruler createLogRuler() {

    Slider.Ruler ruller = _slider.getNewRuler();

    for (double x = Math.floor(_min); x <= Math.ceil(_max); x += 1.) {
      double y = Math.pow(10., x);
      int pos = convertToSlider(y);
      if (pos > -3 && pos < _slider._size + 3) {
        ruller.addLabel(pos, convertToField(y));
        ruller.addMajorTick(pos);
      }
      for (int i=2; i<10; i++) {
        double z = y * i;
        pos = convertToSlider(z);
        if (pos > -3 && pos < _slider._size + 3) {
          ruller.addMinorTick(pos);
          if (i == 5) ruller.addMajorTick(pos);
        }
      }
    }

    return ruller;
  }

// -----------------------------------------------------------------------------
}
