// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import org.freehep.util.Value;

/**
 * {@link Condition} based on comparison with integer range.
 *
 * @author Mark Donszelmann
 * @author Dmitry Onoprienko
 * @version $Id: IntegerCondition.java 13824 2011-05-04 18:30:02Z onoprien $
 */
public class IntegerCondition extends AbstractCondition {

// -- Private parts : ----------------------------------------------------------

  int _min;
  int _max;

// -- Construction and initialization : ----------------------------------------

  public IntegerCondition(ConditionDefinition definition) {
    this(definition, false, false, 0, 0);
  }

  public IntegerCondition(ConditionDefinition definition, boolean enabled, boolean inverted) {
    this(definition, enabled, inverted, 0, 0);
  }

  public IntegerCondition(ConditionDefinition definition, int min, int max) {
    this(definition, false, false, min, max);
  }

  public IntegerCondition(ConditionDefinition definition, boolean enabled, boolean inverted, int min, int max) {
    super(definition, enabled, inverted);
    _min = min;
    _max = max;
  }


// -- Applying condition : -----------------------------------------------------

  /**
   * Returns <tt>true</tt> is the value satisfies this condition.
   * If the condition is not inverted, <tt>minimum &le; value &le; maximum</tt> must be satisfied.
   * If the condition is inverted, <tt>value &lt; minimum</tt> or <tt>maximum &lt; value</tt> must be satisfied.
   */
  public boolean pass(Value value) {
    if (_set) {
      int x = value.getInt();
      return _inverted ? (x < _min || x > _max) :  (x >= _min && x <= _max);
    } else {
      return !_inverted;
    }
  }

// -- Setters : ----------------------------------------------------------------

  /** Sets low bound of the acceptance range. */
  public void setMinimum(int min) {
    _min = min;
    _set = true;
    fireStateChanged();
  }

  /** Sets high bound of the acceptance range. */
  public void setMaximum(int max) {
    _max = max;
    _set = true;
    fireStateChanged();
  }

  /** Sets the acceptance range. */
  public void setRange(int min, int max) {
    _min = min;
    _max = max;
    _set = true;
    fireStateChanged();
  }

  /**
   * Sets the acceptance range.
   * The actual type of the argument should be int[], otherwise <tt>ClassCastException</tt> is thrown.
   */
  public void set(Object value) {
    int[] range = (int[]) value;
    setRange(range[0], range[1]);
  }

// -- Getters : ----------------------------------------------------------------

  /** Returns status string describing what attribute values satisfy this condition in its current state. */
  public String getStatus() {
    if (!_set) return _inverted ? "none" : "all";
    if (_inverted) {
      return "x < "+ _min +" or "+ _max +" < x";
    } else {
      return (_min == _max) ? "x = "+ _min : _min +" <= x <= "+ _max;
    }
  }

  /** Returns low bound of the acceptance range. */
  public int getMinimum() {
    return _min;
  }

  /** Returns high bound of the acceptance range. */
  public int getMaximum() {
    return _max;
  }

  /** Returns the acceptance range. */
  public int[] get() {
    return new int[] {_min ,_max};
  }

// -----------------------------------------------------------------------------
}
