// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import org.freehep.util.Value;

/**
 * {@link Condition} based on comparison with <tt>long</tt> range.
 *
 * @author Mark Donszelmann
 * @author Dmitry Onoprienko
 * @version $Id: LongCondition.java 13824 2011-05-04 18:30:02Z onoprien $
 */
public class LongCondition extends AbstractCondition {

// -- Private parts : ----------------------------------------------------------

  long _min;
  long _max;

// -- Construction and initialization : ----------------------------------------

  public LongCondition(ConditionDefinition definition) {
    this(definition, false, false, 0L, 0L);
  }

  public LongCondition(ConditionDefinition definition, boolean enabled, boolean inverted) {
    this(definition, enabled, inverted, 0L, 0L);
  }

  public LongCondition(ConditionDefinition definition, long min, long max) {
    this(definition, false, false, min, max);
  }

  public LongCondition(ConditionDefinition definition, boolean enabled, boolean inverted, long min, long max) {
    super(definition, enabled, inverted);
    _min = min;
    _max = max;
  }


// -- Applying condition : -----------------------------------------------------

  /**
   * Returns <tt>true</tt> is the value satisfies this condition.
   * If the condition is not inverted, <tt>minimum &le; value &le; maximum</tt> must be satisfied.
   * If the condition is inverted, <tt>value &lt; minimum</tt> or <tt>maximum &lt; value</tt> must be satisfied.
   */
  public boolean pass(Value value) {
    if (_set) {
      long x = value.getLong();
      return _inverted ? (x < _min || x > _max) :  (x >= _min && x <= _max);
    } else {
      return !_inverted;
    }
  }

// -- Setters : ----------------------------------------------------------------

  /** Sets low bound of the acceptance range. */
  public void setMinimum(long min) {
    _min = min;
    _set = true;
    fireStateChanged();
  }

  /** Sets high bound of the acceptance range. */
  public void setMaximum(long max) {
    _max = max;
    _set = true;
    fireStateChanged();
  }

  /** Sets the acceptance range. */
  public void setRange(long min, long max) {
    _min = min;
    _max = max;
    _set = true;
    fireStateChanged();
  }

  /**
   * Sets the acceptance range.
   * The actual type of the argument should be long[], otherwise <tt>ClassCastException</tt> is thrown.
   */
  public void set(Object value) {
    long[] range = (long[]) value;
    setRange(range[0], range[1]);
  }

// -- Getters : ----------------------------------------------------------------

  /** Returns status string describing what attribute values satisfy this condition in its current state. */
  public String getStatus() {
    if (!_set) return _inverted ? "none" : "all";
    if (_inverted) {
      return "x < "+ _min +" or "+ _max +" < x";
    } else {
      return (_min == _max) ? "x = "+ _min : _min +" <= x <= "+ _max;
    }
  }

  /** Returns the acceptance range. */
  public long[] get() {
    return new long[] {_min ,_max};
  }

// -----------------------------------------------------------------------------
}
