// Copyright 2003-2004, FreeHEP.
package hep.wired.edit;

import hep.wired.feature.Scaleable;
import hep.wired.feature.Scaleable2D;
import hep.wired.feature.Scaleable3D;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;

import java.awt.Component;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.util.logging.Logger;

import javax.swing.undo.UndoableEdit;

/**
 * Scales a plot in 2D or 3D.
 * 
 * @author Mark Donszelmann
 * @version $Id: Scale.java 8598 2006-08-14 20:38:02Z duns $
 */
public class Scale extends AnimatedWiredEdit implements GraphicsPanelEdit {

    private static Logger logger = Logger.getLogger(Scale.class.getPackage()
            .getName());

    private double sx, sy, sz;

    /**
     * Creates a scale edit with no scaling done.
     */
    public Scale() {
        this(1);
    }

    /**
     * Creates a scale edit with uniform given scale factor.
     */
    public Scale(double s) {
        this(s, s);
    }

    /**
     * Creates a scale edit with given scale factors.
     */
    public Scale(double sx, double sy) {
        this(sx, sy, 1);
    }

    /**
     * Creates a scale edit with given scale factors.
     */
    public Scale(double sx, double sy, double sz) {
        this(sx, sy, sz, null, 0);
    }

    /**
     * Creates a scale edit with given scale factor, animation shape and number
     * of frames.
     */
    public Scale(double sx, double sy, double sz, Shape shape, int frames) {
        super(shape, frames);
        this.sx = sx;
        this.sy = sy;
        this.sz = sz;
    }

    public WiredEdit copy(RecordPlot plot) {
        WiredEdit copy = new Scale(sx, sy, sz, getShape(), getFrames());
        copy.setRecordPlot(plot);
        return copy;
    }

    protected Shape[] getShapes(Shape shape, int steps) {
        if ((shape == null) || (steps <= 1))
            return null;

        Shape[] shapes = new Shape[steps];

        // calculate scale increments
        double dsx = (sx - 1.0) / (steps - 1);
        double dsy = (sy - 1.0) / (steps - 1);

        // calculate window offset
        double w = getRecordPlot().getWidth() / 2.0;
        double h = getRecordPlot().getHeight() / 2.0;

        // calculate steps
        for (int i = 0; i < steps; i++) {
            AffineTransform transform = new AffineTransform();
            // put in the center
            transform.translate(w, h);
            // add our increment
            transform.scale(1 + (i * dsx), 1 + (i * dsy));
            // put back
            transform.translate(-w, -h);
            // transform
            shapes[i] = transform.createTransformedShape(shape);
        }
        return shapes;
    }

    public Shape createTransformedShape(Component component, Shape shape) {

        double w = component.getWidth() / 2.0;
        double h = component.getHeight() / 2.0;
        AffineTransform transform = new AffineTransform();
        transform.translate(w, h);
        transform.scale(sx, sy);
        transform.translate(-w, -h);
        return transform.createTransformedShape(shape);
    }

    public String getPresentationName() {
        StringBuffer s = new StringBuffer();
        s.append(toString());
        s.append(" (");
        s.append(format.format(sx));
        if ((sy != 1) || (sz != 1)) {
            s.append(", ");
            s.append(format.format(sy));
        }
        if (sz != 1) {
            s.append(", ");
            s.append(format.format(sz));
        }
        s.append(")");
        return s.toString();
    }

    public String toString() {
        return "Scale";
    }

    public boolean addEdit(UndoableEdit edit) {
        if (edit instanceof Scale) {
            Scale scale = (Scale) edit;
            sx *= scale.sx;
            sy *= scale.sy;
            sz *= scale.sz;
            return true;
        }
        return false;
    }

    public boolean isSupportedBy(GraphicsPanel p) {
        return p.supports(Scaleable.class) || p.supports(Scaleable2D.class)
                || p.supports(Scaleable3D.class);
    }

    protected void redoEdit() {
        GraphicsPanel panel = getRecordPlot().getGraphicsPanel();
        Scaleable3D scaleable3D = (Scaleable3D) panel
                .getFeature(Scaleable3D.class);
        if (scaleable3D != null) {
            scaleable3D.scale(sx, sy, sz);
        } else {
            Scaleable2D scaleable2D = (Scaleable2D) panel
                    .getFeature(Scaleable2D.class);
            if (scaleable2D != null) {
                scaleable2D.scale(sx, sy);
            } else {
                Scaleable scaleable = (Scaleable) panel
                        .getFeature(Scaleable.class);
                if (scaleable != null) {
                    scaleable.scale(sx);
                } else {
                    System.err.println("Could not redo: " + this);
                    return;
                }
            }
        }
        getRecordPlot().repaint();
        logger.finer(toString());
    }

    protected void undoEdit() {
        GraphicsPanel panel = getRecordPlot().getGraphicsPanel();
        Scaleable3D scaleable3D = (Scaleable3D) panel
                .getFeature(Scaleable3D.class);
        if (scaleable3D != null) {
            scaleable3D.scale(1 / sx, 1 / sy, 1 / sz);
        } else {
            Scaleable2D scaleable2D = (Scaleable2D) panel
                    .getFeature(Scaleable2D.class);
            if (scaleable2D != null) {
                scaleable2D.scale(1 / sx, 1 / sy);
            } else {
                Scaleable scaleable = (Scaleable) panel
                        .getFeature(Scaleable.class);
                if (scaleable != null) {
                    scaleable.scale(1 / sx);
                } else {
                    System.err.println("Could not undo: " + this);
                    return;
                }
            }
        }
        getRecordPlot().repaint();
        logger.finer(toString());
    }
}
