// Copyright 2004-2005, FreeHEP.
package hep.wired.edit;

import hep.wired.feature.Shearable2D;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;

import java.awt.Component;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.util.logging.Logger;

import javax.swing.undo.UndoableEdit;

/**
 * Shears a plot in 2D.
 * 
 * @author Mark Donszelmann
 * @version $Id: Shear.java 8598 2006-08-14 20:38:02Z duns $
 */
public class Shear extends AnimatedWiredEdit implements GraphicsPanelEdit {

    private static Logger logger = Logger.getLogger(Shear.class.getPackage()
            .getName());

    private double shx, shy;

    /**
     * Creates a shear edit with no shearing.
     */
    public Shear() {
        this(0, 0);
    }

    /**
     * Creates a shear edit with given shearing.
     */
    public Shear(double shx, double shy) {
        this(shx, shy, null, 0);
    }

    /**
     * Creates a shear edit with given shearing, animation shape and number of
     * frames.
     */
    public Shear(double shx, double shy, Shape shape, int frames) {
        super(shape, frames);
        this.shx = shx;
        this.shy = shy;
    }

    public WiredEdit copy(RecordPlot plot) {
        WiredEdit copy = new Shear(shx, shy, getShape(), getFrames());
        copy.setRecordPlot(plot);
        return copy;
    }

    protected Shape[] getShapes(Shape shape, int steps) {
        if ((shape == null) || (steps <= 1))
            return null;

        Shape[] shapes = new Shape[steps];

        // calculate shear increments
        double dshx = shx / (steps - 1);
        double dshy = shy / (steps - 1);

        // calculate window offset
        double w = getRecordPlot().getWidth() / 2.0;
        double h = getRecordPlot().getHeight() / 2.0;

        // calculate steps
        for (int i = 0; i < steps; i++) {
            AffineTransform transform = new AffineTransform();
            // put in the center
            transform.translate(w, h);
            // add our shear increment
            transform.shear(-i * dshx, -i * dshy);
            // put back
            transform.translate(-w, -h);
            // transform
            shapes[i] = transform.createTransformedShape(shape);
        }
        return shapes;
    }

    public Shape createTransformedShape(Component component, Shape shape) {
        double w = component.getWidth() / 2.0;
        double h = component.getHeight() / 2.0;

        AffineTransform transform = new AffineTransform();
        transform.translate(w, h);
        transform.shear(-shx, -shy);
        transform.translate(-w, -h);
        return transform.createTransformedShape(shape);
    }

    public String getPresentationName() {
        return toString() + " (" + format.format(shx) + ", "
                + format.format(shy) + ")";
    }

    public String toString() {
        return "Shear";
    }

    public boolean addEdit(UndoableEdit edit) {
        if (edit instanceof Shear) {
            Shear shear = (Shear) edit;
            shx *= shear.shx;
            shy *= shear.shy;
            return true;
        }
        return false;
    }

    public boolean isSupportedBy(GraphicsPanel p) {
        return p.supports(Shearable2D.class);
    }

    protected void redoEdit() {
        Shearable2D shearable = (Shearable2D) getRecordPlot()
                .getGraphicsPanel().getFeature(Shearable2D.class);
        if (shearable == null) {
            System.err.println("Could not redo: " + this);
            return;
        }
        shearable.shear(shx, shy);
        getRecordPlot().repaint();
        logger.finer(toString());
    }

    protected void undoEdit() {
        Shearable2D shearable = (Shearable2D) getRecordPlot()
                .getGraphicsPanel().getFeature(Shearable2D.class);
        if (shearable == null) {
            System.err.println("Could not undo: " + this);
            return;
        }
        shearable.shear((shx == 0) ? 0 : 1 / shx, (shy == 0) ? 0 : 1 / shy);
        getRecordPlot().repaint();
        logger.finer(toString());
    }
}
