// Copyright 2004-2005, FreeHEP.
package hep.wired.edit;

import hep.wired.feature.Transformable2D;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;

import java.awt.Component;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.util.logging.Logger;

import javax.swing.undo.UndoableEdit;

/**
 * Transforms a plot in 2D.
 * 
 * @author Mark Donszelmann
 * @version $Id: Transform2D.java 8598 2006-08-14 20:38:02Z duns $
 */
public class Transform2D extends AnimatedWiredEdit implements GraphicsPanelEdit {

    private static Logger logger = Logger.getLogger(Transform2D.class
            .getPackage().getName());

    private double sx, shy, shx, sy, tx, ty;

    /**
     * Creates a transform edit with no changes.
     */
    public Transform2D() {
        this(1, 0, 0, 1, 0, 0);
    }

    /**
     * Creates a transform edit with given scaling, shearing and translation.
     */
    public Transform2D(double sx, double shy, double shx, double sy, double tx,
            double ty) {
        this(sx, shy, shx, sy, tx, ty, null, 0);
    }

    /**
     * Creates a transform edit with given scaling, shearing, translation,
     * animation shape and number of frames.
     */
    public Transform2D(double sx, double shy, double shx, double sy, double tx,
            double ty, Shape shape, int frames) {
        super(shape, frames);
        this.sx = sx;
        this.shy = shy;
        this.shx = shx;
        this.sy = sy;
        this.tx = tx;
        this.ty = ty;
    }

    public WiredEdit copy(RecordPlot plot) {
        WiredEdit copy = new Transform2D(sx, shy, shx, sy, tx, ty, getShape(),
                getFrames());
        copy.setRecordPlot(plot);
        return copy;
    }

    protected Shape[] getShapes(Shape shape, int steps) {
        if ((shape == null) || (steps <= 1))
            return null;

        Shape[] shapes = new Shape[steps];

        // calculate scale increments
        double dsx = (sx - 1.0) / (steps - 1);
        double dsy = (sy - 1.0) / (steps - 1);

        // calculate shear increments
        double dshx = shx / (steps - 1);
        double dshy = shy / (steps - 1);

        // calculate translate increments
        double dtx = tx / (steps - 1);
        double dty = ty / (steps - 1);

        // calculate window offset
        double w = getRecordPlot().getWidth() / 2.0;
        double h = getRecordPlot().getHeight() / 2.0;

        // calculate steps
        for (int i = 0; i < steps; i++) {
            AffineTransform transform = new AffineTransform();
            // put in the center
            transform.translate(w, h);
            // add our scale increment
            transform.scale(1 + (i * dsx), 1 + (i * dsy));
            // add our shear increment
            transform.shear(-i * dshx, -i * dshy);
            // add our translate increment
            transform.translate(-i * dtx, i * dty);
            // put back
            transform.translate(-w, -h);
            // transform
            shapes[i] = transform.createTransformedShape(shape);
        }
        return shapes;
    }

    public Shape createTransformedShape(Component component, Shape shape) {
        double w = component.getWidth() / 2.0;
        double h = component.getHeight() / 2.0;

        AffineTransform transform = new AffineTransform();
        transform.translate(w, h);
        transform.scale(sx, sy);
        transform.shear(-shx, -shy);
        transform.translate(-tx, ty);
        transform.translate(-w, -h);
        return transform.createTransformedShape(shape);
    }

    public String getPresentationName() {
        return "Transform2D"; // ("+sx+", "+shy+", "+shx+", "+sy+", "+tx+",
                                // "+ty+")";
    }

    public boolean addEdit(UndoableEdit edit) {
        if (edit instanceof Transform2D) {
            // FIXME WIRED-248 check this, one could multiply transforms
            return false;
        }
        return false;
    }

    public boolean isSupportedBy(GraphicsPanel p) {
        return p.supports(Transformable2D.class);
    }

    protected void redoEdit() {
        Transformable2D transformable2D = (Transformable2D) getRecordPlot()
                .getGraphicsPanel().getFeature(Transformable2D.class);
        if (transformable2D == null) {
            System.err.println("Could not redo: " + this);
            return;
        }

        transformable2D.scale(sx, sy);
        transformable2D.shear(shx, shy);
        transformable2D.translate(tx, ty);

        getRecordPlot().repaint();
        logger.finer(toString());
    }

    protected void undoEdit() {
        Transformable2D transformable2D = (Transformable2D) getRecordPlot()
                .getGraphicsPanel().getFeature(Transformable2D.class);
        if (transformable2D == null) {
            System.err.println("Could not undo: " + this);
            return;
        }

        transformable2D.translate(-tx, -ty);
        transformable2D.shear(-shx, -shy);
        transformable2D.scale(1 / sx, 1 / sy);

        getRecordPlot().repaint();
        logger.finer(toString());
    }

    public String toString() {
        return super.toString() + "; (" + sx + ", " + shy + ", " + shx + ", "
                + sy + ", " + tx + ", " + ty + ")";
    }
}
