// Copyright 2003-2005, FreeHEP.
package hep.wired.interaction;

import hep.wired.edit.Rotate;
import hep.wired.edit.Scale;
import hep.wired.edit.Transform2D;
import hep.wired.edit.Translate;
import hep.wired.edit.WiredCompoundEdit;
import hep.wired.feature.Translateable;
import hep.wired.image.WiredBaseImage;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;

import java.awt.Component;
import java.awt.Shape;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;

import javax.swing.Icon;

import org.freehep.application.Application;
import org.freehep.util.DiscreteAngle;

/**
 * Drag/click mouse to define an (angled) rectangle to scale, rotate and
 * translate plot.
 * 
 * Implements: 1. CMCMC - Click, Move, Click, Move, Click 2. CMD - Click, Move,
 * Drag 3. DMC - Drag, Move, Click 4. DMD - Drag, Move, Drag
 * 
 * @author Mark Donszelmann
 * @version $Id: DragAngledRectangleToScale.java 8598 2006-08-14 20:38:02Z duns $
 */
public class DragAngledRectangleToScale extends DefaultInteractionHandler {

    private static final String name = "Combined Transformation";

    private static final int cursorSize = 32;

    private boolean fixedRatio = true;

    private int n;

    private int[] x = new int[3];

    private int[] y = new int[3];

    private int xp, yp;

    private DiscreteAngle angle;

    private GeneralPath path;
    
    /**
     * Create a angled rectangle handler to generate scal edits.
     */
    public DragAngledRectangleToScale() {
        super("Drag Angled Rectangle to Scale");
    }

    public Icon getIcon(int size) {
        return WiredBaseImage.getIcon("AngledRectangle%w", size);
    }

    public String getDescription() {
        return "Click to define first point of angled rectangle.";
    }

    public boolean isSupportedBy(GraphicsPanel panel) {
        return new Scale().isSupportedBy(panel)
                && new Transform2D().isSupportedBy(panel)
                && new Translate().isSupportedBy(panel)
                && new Rotate().isSupportedBy(panel);
    }

    public void changeCursor(RecordPlot plot, InputEvent event) {
        if ((n == 2) && !event.isControlDown()) {
            // calculate cursor
            if (x[2] == x[1]) {
                plot.setCursor(WiredBaseImage.getBestCursor("YSkewCursor%w",
                        cursorSize, cursorSize));
            } else if (y[2] == y[1]) {
                plot.setCursor(WiredBaseImage.getBestCursor("XSkewCursor%w",
                        cursorSize, cursorSize));
            } else {
                plot.setCursor(WiredBaseImage.getBestCursor(
                        "AngledRectangleCursor%w", cursorSize, cursorSize));
            }
        } else {
            plot.setCursor(WiredBaseImage.getBestCursor(
                    "AngledRectangleCursor%w", cursorSize, cursorSize));
        }
    }

    public void reset(RecordPlot plot, InputEvent event) {
        n = 0;
        plot.drawShape(null);
        changeCursor(plot, event);
        Application.getApplication().setStatusMessage(getDescription());
    }

    public void mouseEntered(RecordPlot plot, MouseEvent event) {
        plot.requestFocusInWindow();
        changeCursor(plot, event);
    }

    public void mouseButton1Clicked(RecordPlot plot, MouseEvent event) {
        definePoint(plot, event);
        pinPoint(plot, event);
    }

    public void mouseButton1DragStarted(RecordPlot plot, MouseEvent event) {
        definePoint(plot, event);
        // last point is defined by end of the drag
        if (n < 2) {
            pinPoint(plot, event);
        }
    }

    public void mouseButton1Dragged(RecordPlot plot, MouseEvent event) {
        // keep point moving
        if (n > 0) {
            definePoint(plot, event);
        }
    }

    public void mouseButton1DragEnded(RecordPlot plot, MouseEvent event) {
        definePoint(plot, event);
        pinPoint(plot, event);
    }

    public void mouseMoved(RecordPlot plot, MouseEvent event) {
        mouseButton1Dragged(plot, event);
    }

    public boolean shiftKeyPressed(RecordPlot plot, KeyEvent event) {
        modifyPoint(plot, event);
        return true;
    }

    public boolean shiftKeyReleased(RecordPlot plot, KeyEvent event) {
        modifyPoint(plot, event);
        return true;
    }

    public boolean ctrlMetaKeyPressed(RecordPlot plot, KeyEvent event) {
        modifyPoint(plot, event);
        return true;
    }

    public boolean ctrlMetaKeyReleased(RecordPlot plot, KeyEvent event) {
        modifyPoint(plot, event);
        return true;
    }

    private void definePoint(RecordPlot plot, MouseEvent event) {
        xp = event.getX();
        yp = event.getY();
        switch (n) {
        case 0:
            x[0] = x[1] = x[2] = xp;
            y[0] = y[1] = y[2] = yp;
            break;
        case 1:
            x[1] = x[2] = xp;
            y[1] = y[2] = yp;
            plot.drawShape(new Line2D.Double(x[0], y[0], x[1], y[1]));
            break;
        case 2:
            x[2] = xp;
            y[2] = yp;
            modifyPoint(plot, event);
            break;
        default:
            finished(plot, event);
            break;
        }
    }

    private void modifyPoint(RecordPlot plot, InputEvent event) {
        if (n != 2)
            return;
        if (!event.isControlDown()) {
            int dx = xp - x[1];
            int dy = yp - y[1];
            // calculate drag angle
            double alpha = (Math.atan2(dy, dx) + 2 * Math.PI) % (2 * Math.PI);

            // calculate discrete angle
            double beta = angle.getAngle(alpha);

            // calculate length of vector along discrete angle
            double cosBeta = Math.cos(beta);
            double sinBeta = Math.sin(beta);
            double h1 = Math.abs((cosBeta < 0.1) ? dx : dx / cosBeta);
            double h2 = Math.abs((sinBeta < 0.1) ? dy : dy / sinBeta);
            double height2 = Math.max(h1, h2);

            // restrict ratio
            if (!event.isShiftDown()) {
                int wx = x[1] - x[0];
                int wy = y[1] - y[0];
                double width = Math.sqrt(wx * wx + wy * wy);
                double ratio = (double) plot.getHeight()
                        / (double) plot.getWidth();
                height2 = ratio * width / 2;
                fixedRatio = true;
            } else {
                fixedRatio = false;
            }

            // calculate point
            x[2] = (int) (height2 * cosBeta + x[1]);
            y[2] = (int) (height2 * sinBeta + y[1]);
        } else {
            x[2] = xp;
            y[2] = yp;
        }

        // define shape
        int dx = x[2] - x[1];
        int dy = y[2] - y[1];
        path = new GeneralPath();
        path.moveTo(x[0], y[0]);
        path.lineTo(x[1], y[1]);
        path.moveTo(x[0] + dx, y[0] + dy);
        path.lineTo(x[1] + dx, y[1] + dy);
        path.lineTo(x[1] - dx, y[1] - dy);
        path.lineTo(x[0] - dx, y[0] - dy);
        path.closePath();
        plot.drawShape(path);
        changeCursor(plot, event);
    }

    private void pinPoint(RecordPlot plot, InputEvent event) {
        switch (n) {
        case 0:
            Application.getApplication().setStatusMessage(
                    "Click to define angle of angled rectangle.");
            angle = new DiscreteAngle();
            angle.addAngle(0);
            angle.addAngle(Math.PI / 2);
            angle.addAngle(Math.PI);
            angle.addAngle(3 * Math.PI / 2);
            angle.addAngle(2 * Math.PI);
            n++;
            break;
        case 1:
            Application
                    .getApplication()
                    .setStatusMessage(
                            "Click to define rectangle, Shift removes ratio constraint, "+(isMac() ? command : "Ctrl")+" removes angle and ratio constraint.");
            double perp = (Math.atan2(y[1] - y[0], x[1] - x[0]) + 3 * Math.PI / 2.0)
                    % (2 * Math.PI);
            /* Double perpAngle1 = */ angle.addAngle(perp);
            /* Double perpAngle2 = */ angle.addAngle((perp + Math.PI) % (2 * Math.PI));
            n++;
            break;
        case 2:
            finished(plot, event);
            n = 0;
            break;
        default:
            // Should never happen.
            break;
        }
    }

    private void finished(RecordPlot plot, InputEvent event) {
        WiredCompoundEdit edit = new WiredCompoundEdit(name);

        int dx1 = x[1] - x[0];
        int dy1 = y[1] - y[0];
        int dx2 = x[2] - x[1];
        int dy2 = y[2] - y[1];
        double w = Math.sqrt(dx1 * dx1 + dy1 * dy1);
        double h = Math.sqrt(dx2 * dx2 + dy2 * dy2) * 2;

        int w2 = plot.getWidth() / 2;
        int h2 = plot.getHeight() / 2;

        if ((w != 0) && (h != 0)) {

            // center
            int dx = dx1 / 2 + x[0] - w2;
            int dy = dy1 / 2 + y[0] - h2;
            Translateable translateable = (Translateable) plot
                    .getGraphicsPanel().getFeature(Translateable.class);
            double t[] = translateable.getModelTranslation(new double[] { -dx,
                    dy, 0 }, plot.getGraphicsPanel().getViewPort());
            Translate translate = new Translate(t[0], t[1], t[2], path, 5);
            Shape shape = translate.createTransformedShape((Component) plot,
                    path);
            edit.addEdit(translate);

            // Rotate
            double alpha = Math.atan2(dy1, dx1);
            if (alpha > Math.PI / 2)
                alpha -= Math.PI;
            if (alpha <= -Math.PI / 2)
                alpha += Math.PI;
            Rotate rotate = new Rotate(alpha, 0, 0, 1, shape, 5);
            shape = rotate.createTransformedShape((Component) plot, shape);
            edit.addEdit(rotate);

            // Shear
            double beta = Math.atan2(dy2, dx2);
            double delta = Math.PI / 2.0 - beta + alpha;
            if (delta > Math.PI / 2)
                delta -= Math.PI;
            if (delta <= -Math.PI / 2)
                delta += Math.PI;
            double shx = Math.tan(delta);
            double esy = h / (h * Math.cos(delta));
            Transform2D transform = new Transform2D(1, 0, shx, esy, 0, 0,
                    shape, 5);
            shape = transform.createTransformedShape((Component) plot, shape);
            edit.addEdit(transform);

            // Zoom
            double sx = plot.getWidth() / (w + 1);
            double sy = fixedRatio ? sx : plot.getHeight() / (h + 1);
            double sz = fixedRatio ? sx : Math.min(sx, sy);
            edit.addEdit(new Scale(sx, sy, sz, shape, 5));

            // send
            edit.end();
            plot.postEdit(edit);
        }

        reset(plot, event);
    }

    public String toString() {
        return "Drag angled rectangle to Scale";
    }
}
