// Copyright 2003-2006, FreeHEP.
package hep.wired.interaction;

import hep.wired.edit.Scale;
import hep.wired.edit.Translate;
import hep.wired.edit.WiredCompoundEdit;
import hep.wired.feature.Translateable;
import hep.wired.image.WiredBaseImage;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;

import java.awt.Component;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;

import javax.swing.Icon;

import org.freehep.application.Application;

/**
 * Drag/click to define a rectangle to scale and translate the plot.
 * 
 * @author Mark Donszelmann
 * @version $Id: DragRectangleToScale.java 13826 2011-05-05 03:05:38Z onoprien $
 */
public class DragRectangleToScale extends DefaultInteractionHandler {

    private static final String name = "Combined Translate-Scale";

    private static final int cursorSize = 32;

    private static final double fixedScale = Math.sqrt(2);

    private boolean fixedRatio = true;

    private int xp, yp;

    private int xd, yd;

    private Rectangle rectangle;

    /**
     * Create a rectangle handler to generate scale edits.
     */
    public DragRectangleToScale() {
        super("Drag Rectangle to Scale");
    }

    public Icon getIcon(int size) {
        return WiredBaseImage.getIcon("RectangleScale%w", size);
    }

    public String getDescription() {
        return "Click center-zoom-in, Shift-Click to zoom-out-displace, Drag rectangle to zoom.";
    }

    public boolean isSupportedBy(GraphicsPanel panel) {
        return new Translate().isSupportedBy(panel)
                && new Scale().isSupportedBy(panel);
    }

    public void changeCursor(RecordPlot plot, InputEvent event) {
        if (event.getID() == MouseEvent.MOUSE_DRAGGED) {
            plot.setCursor(WiredBaseImage.getBestCursor(
                    "RectangleScaleCursor%w", cursorSize, cursorSize));
        } else if (event.isShiftDown()) {
            plot.setCursor(WiredBaseImage.getBestCursor(
                    "RectangleScaleMinusCursor%w", cursorSize, cursorSize));
        } else {
            plot.setCursor(WiredBaseImage.getBestCursor(
                    "RectangleScalePlusCursor%w", cursorSize, cursorSize));
        }
    }

    public void reset(RecordPlot plot, InputEvent event) {
        rectangle = null;
        plot.drawShape(null);
        Application.getApplication().setStatusMessage(getDescription());
    }

    public void mouseEntered(RecordPlot plot, MouseEvent event) {
        plot.requestFocusInWindow();
        changeCursor(plot, event);
    }

    /**
     * Store starting point of rectangle.
     */
    public void mouseButton1DragStarted(RecordPlot plot, MouseEvent event) {
        xp = event.getX();
        yp = event.getY();
        changeCursor(plot, event);
        Application.getApplication().setStatusMessage(
                "Drag rectangle to zoom, Shift removes ratio constraint.");
    }

    /**
     * Update moving point of rectangle and draw.
     */
    public void mouseButton1Dragged(RecordPlot plot, MouseEvent event) {
        xd = event.getX();
        yd = event.getY();
        defineRectangle(plot, event);
    }

    /**
     * Update end point of rectangle, translate and scale, and remove rectangle.
     */
    public void mouseButton1DragEnded(RecordPlot plot, MouseEvent event) {
        translateAndScale(plot, rectangle, fixedRatio);
        reset(plot, event);
    }

    // FIXME, WIRED-377, separate this from the selection
    public static void translateAndScale(RecordPlot plot,
            Rectangle2D rectangle, boolean fixedRatio) {
        double du = rectangle.getX() + rectangle.getWidth() / 2
                - plot.getWidth() / 2;
        double dv = rectangle.getY() + rectangle.getHeight() / 2
                - plot.getHeight() / 2;

        WiredCompoundEdit edit = new WiredCompoundEdit(name);

        // center
        Translateable translateable = (Translateable) plot.getGraphicsPanel()
                .getFeature(Translateable.class);
        double t[] = translateable.getModelTranslation(new double[] { -du, dv,
                0 }, plot.getGraphicsPanel().getViewPort());
        Translate translate = new Translate(t[0], t[1], t[2], rectangle, 5);
        Shape shape = translate.createTransformedShape((Component) plot,
                rectangle);
        edit.addEdit(translate);

        // zoom
        double sx = plot.getWidth() / (rectangle.getWidth() + 1);
        double sy = fixedRatio ? sx : plot.getHeight()
                / (rectangle.getHeight() + 1);
        double sz = fixedRatio ? sx : Math.min(sx, sy);
        edit.addEdit(new Scale(sx, sy, sz, shape, 5));

        edit.end();
        plot.postEdit(edit);
    }

    public void mouseButton1Released(RecordPlot plot, MouseEvent event) {
        changeCursor(plot, event);
    }

    public boolean shiftKeyPressed(RecordPlot plot, KeyEvent event) {
        defineRectangle(plot, event);
        return true;
    }

    public boolean shiftKeyReleased(RecordPlot plot, KeyEvent event) {
        defineRectangle(plot, event);
        return true;
    }

    public boolean altKeyPressed(RecordPlot plot, KeyEvent event) {
        changeCursor(plot, event);
        return true;
    }

    public boolean altKeyReleased(RecordPlot plot, KeyEvent event) {
        changeCursor(plot, event);
        return true;
    }

    private void defineRectangle(RecordPlot plot, InputEvent event) {
        // no need to define if we pressed a key and rectangle was defined
        // before
        if ((event instanceof KeyEvent) && (rectangle == null))
            return;

        int dx = xd - xp;
        int dy = yd - yp;

        // restrict rectangle for plot ratio when shift is not-pressed
        if (!event.isShiftDown()) {
            double ratio = (double) plot.getHeight() / (double) plot.getWidth();
            if (Math.abs(dy) < Math.abs(dx) * ratio) {
                dy = (int) (Math.abs(dx) * ((dy < 0) ? -1 : 1) * ratio);
            } else {
                dx = (int) (Math.abs(dy) * ((dx < 0) ? -1 : 1) / ratio);
            }
            fixedRatio = true;
        } else {
            fixedRatio = false;
        }
        rectangle = new Rectangle(Math.min(xp, xp + dx), Math.min(yp, yp + dy),
                Math.abs(dx), Math.abs(dy));
        plot.drawShape(rectangle);
    }

    /**
     * Translate point to center and zoom in (or out if alt key pressed).
     */
    public void mouseButton1Clicked(RecordPlot plot, MouseEvent event) {

        int w2 = plot.getWidth() / 2;
        int h2 = plot.getHeight() / 2;
        int du = event.getX() - w2;
        int dv = event.getY() - h2;

        WiredCompoundEdit edit = new WiredCompoundEdit(name);

        if (event.isShiftDown()) {
            Shape shape = new Rectangle(0, 0, 2 * w2, 2 * h2);

            // zoom out
            Scale scale = new Scale(1 / fixedScale, 1 / fixedScale,
                    1 / fixedScale, shape, 5);
            shape = scale.createTransformedShape((Component) plot, shape);
            edit.addEdit(scale);

            // offset from center
            Translateable translateable = (Translateable) plot
                    .getGraphicsPanel().getFeature(Translateable.class);
            double t[] = translateable.getModelTranslation(new double[] {
                    du * fixedScale, -dv * fixedScale, 0 }, plot
                    .getGraphicsPanel().getViewPort());
            edit.addEdit(new Translate(t[0], t[1], t[2], shape, 5));
        } else {
            int ws2 = (int) (w2 / fixedScale);
            int hs2 = (int) (h2 / fixedScale);
            Shape shape = new Rectangle(event.getX() - ws2, event.getY() - hs2,
                    2 * ws2, 2 * hs2);

            // center
            Translateable translateable = (Translateable) plot
                    .getGraphicsPanel().getFeature(Translateable.class);
            double t[] = translateable.getModelTranslation(new double[] { -du,
                    dv, 0 }, plot.getGraphicsPanel().getViewPort());
            Translate translate = new Translate(t[0], t[1], t[2], shape, 5);
            shape = translate.createTransformedShape((Component) plot, shape);
            edit.addEdit(translate);

            // zoom in
            edit
                    .addEdit(new Scale(fixedScale, fixedScale, fixedScale,
                            shape, 5));
        }
        edit.end();
        plot.postEdit(edit);
    }

    public String toString() {
        return "Drag rectangle to Scale";
    }
}
