package hep.wired.plot;

import hep.wired.edit.GraphicsPanelEdit;
import hep.wired.edit.NoOperation;
import hep.wired.edit.Reset;
import hep.wired.edit.WiredCompoundEdit;
import hep.wired.edit.WiredEdit;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.InteractionHandler;
import hep.wired.services.RecordPlot;
import hep.wired.util.NullEvent;
import hep.wired.util.SwingEventMulticaster;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.Iterator;
import java.util.logging.Logger;

import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.undo.UndoableEdit;
import javax.swing.undo.UndoableEditSupport;

import org.freehep.application.Application;
import org.freehep.graphics2d.VectorGraphics;
import org.freehep.swing.popup.HasPopupItems;
import org.freehep.swing.undo.AnimatedEditSupport;
import org.freehep.swing.undo.DoableEdit;
import org.freehep.xml.io.XMLIO;
import org.freehep.xml.io.XMLIOManager;
import org.jdom.Element;

/**
 * Defines a single event display plot.
 * 
 * @author Mark Donszelmann
 * @version $Id: WiredPlot.java 14746 2013-05-02 18:59:56Z onoprien $
 */
public class WiredPlot extends JPanel implements RecordPlot, HasPopupItems, XMLIO {

// -- Private parts : ----------------------------------------------------------

  private static Logger logger = Logger.getLogger(WiredPlot.class.getPackage().getName());
  private String name;
  private GraphicsPanel graphicsPanel;
  private UndoableEditSupport undoSupport;
  private transient boolean selected;
  private transient boolean needsReset;
  private transient ChangeListener listener;

  // For the interaction handler
  private InteractionHandler handler;
  private transient boolean pressed, dragged, clicked;
  private transient MouseEvent pressEvent;
  private transient MouseListener mouseListener;
  private transient MouseMotionListener mouseMotionListener;
  private transient MouseWheelListener mouseWheelListener;
  private transient KeyListener keyListener;

  // For the feedback shape drawing
  private Color color = Color.white;
  private double lineWidth = 1.0;
  private Color frameColor = Color.black;
  private double frameWidth = 3.0;
  private Shape shape;
  private static boolean mac = System.getProperty("os.name").toLowerCase().indexOf("mac") != -1;


// -- Construction : -----------------------------------------------------------

  // For use in XMLIO restore
  private WiredPlot() {
    this(null, null, false, null);
  }

  /**
   * Creates a plot without a record.
   */
  public WiredPlot(GraphicsPanel graphicsPanel) {
    this("Plot", graphicsPanel);
  }

  /**
   * Creates a plot with default GraphicsMode.
   */
  public WiredPlot(String name, GraphicsPanel graphicsPanel) {
    this(name, graphicsPanel, true);
  }

  /**
   * Creates a plot with given parameters.
   */
  public WiredPlot(String name, GraphicsPanel graphicsPanel, boolean supportUndo) {
    this(name, graphicsPanel, supportUndo, null);
    // NOTE, default interaction handler missing
  }

  // NOTE the InteractionHandler is there but never used, only to make this
  // ctor different from the one above.
  private WiredPlot(String name, GraphicsPanel graphicsPanel, boolean supportUndo, InteractionHandler interactionHandler) {
    super(false);
    this.name = name;
    this.selected = false;
    this.needsReset = true;

    // set internal behaviour
    setLayout(new BorderLayout());
    setOpaque(false);

    // create graphics panel
    if (graphicsPanel != null) {
      setGraphicsPanel(graphicsPanel);
    }

    // undo support
    if (supportUndo) {
      addUndoSupport();
    }

    // set the interaction handler
    if (interactionHandler != null) {
      setInteractionHandler(interactionHandler);
    }
  }

  /**
   * Returns a copy of this plot, attached to the given objects, from the
   * given parameters.
   */
  public RecordPlot copy(String name, Object record, boolean supportUndo) {
    WiredPlot copy = new WiredPlot(name, graphicsPanel.copy(), supportUndo,
            null);

    copy.setInteractionHandler(getInteractionHandler());
    copy.setRecord(record);

    copy.color = color;
    copy.lineWidth = lineWidth;
    copy.frameColor = frameColor;
    copy.frameWidth = frameWidth;
    copy.shape = shape;

    return copy;
  }


// -- Setting components and responding to plugin-level events : ---------------

  /**
   * Sets graphics panel used by this plot.
   * No notifications are sent to any listeners or components, no additional initialization
   * is done on this plot or the graphics panel.
   */
  private void setGraphicsPanel(GraphicsPanel newPanel) {
    if (graphicsPanel != null) remove((Component) graphicsPanel);
    graphicsPanel = newPanel;
    add((Component) graphicsPanel, BorderLayout.CENTER);
  }

  /**
   * Sets interaction handler on this plot.
   */
  public void setInteractionHandler(InteractionHandler interactionHandler) {

    if (selected) {
      if (handler != null) {
        handler.setSelected(this, false);
      }
      if (interactionHandler != null) {
        interactionHandler.setSelected(this, true);
      }
    }

    Component panel = (Component) graphicsPanel;

    if (mouseListener != null) {
      panel.removeMouseListener(mouseListener);
    }
    if (mouseMotionListener != null) {
      panel.removeMouseMotionListener(mouseMotionListener);
    }
    if (mouseWheelListener != null) {
      panel.removeMouseWheelListener(mouseWheelListener);
    }
    if (keyListener != null) {
      panel.removeKeyListener(keyListener);
    }

    handler = interactionHandler;
    // clear current shape
    drawShape(null);
    repaint();

    if (handler == null) {
      setCursor(null);
    } else {
      handler.setSize(this, getWidth(), getHeight());
      handler.changeCursor(this, new NullEvent(this));
      handler.reset(this, new NullEvent(this));

      mouseListener = new MouseListener() {

        public void mousePressed(MouseEvent event) {
          if (listener != null) {
            listener.stateChanged(new ChangeEvent(this));
          }

          if (SwingUtilities.isLeftMouseButton(event)) {
            // System.err.println("Pressed ");
            // System.err.print(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
            event.getComponent().requestFocusInWindow();
            // System.err.println(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
            pressed = true;
            pressEvent = event;
            clicked = true;
            dragged = false;
            handler.mouseButton1Pressed(WiredPlot.this, event);
          }
        }

        public void mouseReleased(MouseEvent event) {
          if (SwingUtilities.isLeftMouseButton(event)) {
            // System.err.print("Released ");
            // System.err.println(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
            if (dragged) {
              handler.mouseButton1DragEnded(WiredPlot.this,
                      event);
              dragged = false;
            }
            handler.mouseButton1Released(WiredPlot.this, event);
            pressed = false;
            pressEvent = null;
          }
        }

        public void mouseClicked(MouseEvent event) {
          if (SwingUtilities.isLeftMouseButton(event)) {
            if (clicked) {
              // System.err.print("Clicked ");
              // System.err.println(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
              handler.mouseButton1Clicked(WiredPlot.this, event);
              clicked = false;
            }
            pressed = false;
            pressEvent = null;
            dragged = false;
          }
        }

        public void mouseEntered(MouseEvent event) {
          // System.err.print("Entered ");
          // System.err.println(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
          event.getComponent().requestFocusInWindow();
          handler.mouseEntered(WiredPlot.this, event);
        }

        public void mouseExited(MouseEvent event) {
          // System.err.print("Exited ");
          // System.err.println(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
          handler.mouseExited(WiredPlot.this, event);
        }
      };
      panel.addMouseListener(mouseListener);

      mouseMotionListener = new MouseMotionListener() {

        public void mouseDragged(MouseEvent event) {
          // System.err.print("Dragged ");
          // System.err.println(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
          if (SwingUtilities.isLeftMouseButton(event)) {
            if (pressed) {
              if (pressEvent != null) {
                dragged = true;
                clicked = false;
                handler.mouseButton1DragStarted(WiredPlot.this,
                        pressEvent);
                pressEvent = null;
              }
              handler.mouseButton1Dragged(WiredPlot.this, event);
            }
          }
        }

        public void mouseMoved(MouseEvent event) {
          // System.err.print("Moved ");
          // System.err.println(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
          handler.mouseMoved(WiredPlot.this, event);
        }
      };
      panel.addMouseMotionListener(mouseMotionListener);

      mouseWheelListener = new MouseWheelListener() {

        public void mouseWheelMoved(MouseWheelEvent event) {
          // System.err.print("WheelMoved ");
          // System.err.println(SwingUtilities.getRoot(event.getComponent()).isFocusOwner());
          handler.mouseWheelMoved(WiredPlot.this, event);
        }
      };
      panel.addMouseWheelListener(mouseWheelListener);

      keyListener = new KeyListener() {

        private boolean altKeyDown, ctrlKeyDown, shiftKeyDown, metaKeyDown, escKeyDown;

        public void keyPressed(KeyEvent event) {
          boolean consume = false;
          switch (event.getKeyCode()) {
            case KeyEvent.VK_ALT:
              if (!altKeyDown) {
                altKeyDown = true;
                consume = handler.altKeyPressed(WiredPlot.this, event);
              }
              break;
            case KeyEvent.VK_CONTROL:
              if (!ctrlKeyDown) {
                ctrlKeyDown = true;
                if (!mac) {
                  consume = handler.ctrlMetaKeyPressed(WiredPlot.this, event);
                }
                if (!consume) {
                  consume = handler.ctrlKeyPressed(WiredPlot.this, event);
                }
              }
              break;
            case KeyEvent.VK_SHIFT:
              if (!shiftKeyDown) {
                shiftKeyDown = true;
                consume = handler.shiftKeyPressed(WiredPlot.this,
                        event);
              }
              break;
            case KeyEvent.VK_META:
              if (!metaKeyDown) {
                metaKeyDown = true;
                if (mac) {
                  consume = handler.ctrlMetaKeyPressed(WiredPlot.this, event);
                }
                if (!consume) {
                  consume = handler.metaKeyPressed(WiredPlot.this,
                          event);
                }
              }
              break;
            case KeyEvent.VK_ESCAPE:
              if (!escKeyDown) {
                escKeyDown = true;
                pressed = false;
                dragged = false;
                handler.reset(WiredPlot.this, event);
                consume = handler.escKeyPressed(WiredPlot.this,
                        event);
              }
              break;
            default:
              consume = handler.otherKeyPressed(WiredPlot.this, event);
              break;
          }
          if (consume) {
            event.consume();
          }
        }

        public void keyReleased(KeyEvent event) {
          boolean consume = false;
          switch (event.getKeyCode()) {
            case KeyEvent.VK_ALT:
              if (altKeyDown) {
                altKeyDown = false;
                consume = handler.altKeyReleased(WiredPlot.this,
                        event);
              }
              break;
            case KeyEvent.VK_CONTROL:
              if (ctrlKeyDown) {
                ctrlKeyDown = false;
                if (!mac) {
                  consume = handler.ctrlMetaKeyReleased(WiredPlot.this, event);
                }
                if (!consume) {
                  consume = handler.ctrlKeyReleased(WiredPlot.this,
                          event);
                }
              }
              break;
            case KeyEvent.VK_SHIFT:
              if (shiftKeyDown) {
                shiftKeyDown = false;
                consume = handler.shiftKeyReleased(WiredPlot.this,
                        event);
              }
              break;
            case KeyEvent.VK_META:
              if (metaKeyDown) {
                metaKeyDown = false;
                if (mac) {
                  consume = handler.ctrlMetaKeyReleased(WiredPlot.this, event);
                }
                if (!consume) {
                  consume = handler.metaKeyReleased(WiredPlot.this,
                          event);
                }
              }
              break;
            case KeyEvent.VK_ESCAPE:
              if (escKeyDown) {
                escKeyDown = false;
                consume = handler.escKeyReleased(WiredPlot.this,
                        event);
              }
              break;
            default:
              consume = handler.otherKeyReleased(WiredPlot.this,
                      event);
              break;
          }
          if (consume) {
            event.consume();
          }
        }

        public void keyTyped(KeyEvent event) {
          boolean consume = handler.keyTyped(WiredPlot.this, event);
          if (consume) {
            event.consume();
          }
        }
      };
      panel.addKeyListener(keyListener);
    }
    panel.requestFocusInWindow();
  }

  /**
   * Called when this plot is selected or deselected.
   * Changes its status and forwards the call to its graphics panel and interaction handler.
   */
  public void setSelected(boolean selected) {
    this.selected = selected;
    graphicsPanel.setSelected(this, selected);
    if (handler != null) handler.setSelected(this, selected);
  }

  /**
   * Sets record to be displayed by this plot.
   * Forwards the call to graphics panel and interaction handler, then repaints itself.
   */
  public void setRecord(Object record) {
    graphicsPanel.setRecord(record);
    if (handler != null)  handler.setRecord(this, record);
    if (needsReset && (record != null)) {
      postEdit(new Reset());
      needsReset = false;
    }
    repaint();
  }


// -- Handling change listeners : ----------------------------------------------

  public void addChangeListener(ChangeListener a) {
    listener = SwingEventMulticaster.add(listener, a);
  }

  public void removeChangeListener(ChangeListener r) {
    listener = SwingEventMulticaster.remove(listener, r);
  }
  
// -- Getters : ----------------------------------------------------------------

  /** Returns the name of the plot. */
  public String getName() {
    return name;
  }

  /** Returns the associated Graphics Panel. */
  public GraphicsPanel getGraphicsPanel() {
    return graphicsPanel;
  }

  /** Return the current interaction handler. */
  public InteractionHandler getInteractionHandler() {
    return handler;
  }

  /** Returns the undo support object, or null if undo was not switched on. */
  public UndoableEditSupport getUndoableEditSupport() {
    return undoSupport;
  }

  /** Returns current record object. */
  public Object getRecord() {
    return graphicsPanel.getRecord();
  }

  /** Returns true if this plot (and its projection) supports the specified interaction handler. */
  public boolean supports(InteractionHandler interactionHandler) {
    return interactionHandler != null ? interactionHandler.isSupportedBy(getGraphicsPanel()) : true;
  }


// -- Edits/Painting : ---------------------------------------------------------

  private void addUndoSupport() {
    undoSupport = new AnimatedEditSupport(this);
  }

  public boolean requestFocusInWindow() {
    return ((Component) graphicsPanel).requestFocusInWindow();
  }

  public void repaint() {
    super.repaint();
    if (graphicsPanel != null) {
      ((Component) graphicsPanel).repaint();
    }
  }

  // NOTE, to make sure calculations can be done inside setBounds in either
  // InteractionPanel or
  // graphicsPanel, we resize the graphicsPanel and interactionPanel
  // immediately.
  public void setBounds(int x, int y, int width, int height) {
    ((Component) graphicsPanel).setBounds(x, y, width, height);
    if (handler != null) {
      handler.setSize(this, width, height);
    }
    super.setBounds(x, y, width, height);
  }

  public String toString() {
    return (getName() == null) ? "WiredPlot@" + hashCode() : "WiredPlot: "
            + getName();
  }

  //
  // UndoableEditProcessor interface
  //
  /**
   * Returns true if given edit is supported, which may depend on the current
   * graphics panel.
   */
  public boolean supports(UndoableEdit edit) {
    if (edit instanceof NoOperation) {
      return true;
    }

    if (edit instanceof GraphicsPanelEdit) {
      GraphicsPanelEdit graphicsPanelEdit = (GraphicsPanelEdit) edit;
      return graphicsPanelEdit.isSupportedBy(getGraphicsPanel());
    }

    return false;
  }

  /** Handles an undoable edit, please use postEdit(WiredEdit). */
  public void postEdit(UndoableEdit edit) {
    if (edit instanceof DoableEdit) {
      postEdit((DoableEdit) edit);
      return;
    }
    logger.warning("Edit ignored: " + edit);
  }

  /** Handles an doable edit, please use postEdit(WiredEdit). */
  public void postEdit(DoableEdit edit) {
    if (edit instanceof WiredEdit) {
      postEdit((WiredEdit) edit);
      return;
    }
    if (edit instanceof WiredCompoundEdit) {
      postEdit((WiredCompoundEdit) edit);
      return;
    }
    logger.warning("Edit should be Wired(Compound)Edit...: " + edit);
    edit.redo();
    if (edit.isSignificant()) {
      if (undoSupport != null) {
        undoSupport.postEdit(edit);
      }
    }
  }

  /**
   * Handles a Wired Edit. You may post the same edit to different plots,
   * since edits are copied and afterwards attached to the plot. Any edits
   * posted can be undone, if undoable support was switched on at creation
   * time of the plot. Each plot has its own undo support.
   */
  public void postEdit(WiredEdit edit) {
    WiredEdit wiredEdit = edit.copy(this);

    // logger.info(wiredEdit.toString());
    wiredEdit.redo();
    if (wiredEdit.isSignificant()) {
      if (undoSupport != null) {
        undoSupport.postEdit(wiredEdit);
      }
    }
  }

  /**
   * Handles a Wired Compound Edit. You may post the same edit to different
   * plots, since edits are copied (recursively) and afterwards attached to
   * the plot. Any edits posted can be undone, if undoable support was
   * switched on at creation time of the plot. Each plot has its own undo
   * support.
   */
  public void postEdit(WiredCompoundEdit edit) {
    WiredCompoundEdit wiredEdit = edit.copy(this);

    // logger.info(wiredEdit.toString());
    wiredEdit.redo();
    if (wiredEdit.isSignificant()) {
      if (undoSupport != null) {
        undoSupport.postEdit(wiredEdit);
      }
    }
  }

  /**
   * If multiple edits need to count as one major update, call this method to
   * start a compound edit.
   */
  public void beginUpdate() {
    if (undoSupport != null) {
      undoSupport.beginUpdate();
    }
  }

  /**
   * If multiple edits need to count as one major update, call this method to
   * end a compound edit.
   */
  public void endUpdate() {
    if (undoSupport != null) {
      undoSupport.endUpdate();
    }
  }

  public void drawShape(Shape shape) {
    this.shape = shape;
    // only recopy the graphicspanel, repaint the shape
    super.repaint();
  }

  public void paint(Graphics g) {
    super.paint(g);

    // Not needed
    // if (graphicsPanel != null) ((Component)graphicsPanel).paint(g);

    if (shape != null) {
      VectorGraphics graphics = VectorGraphics.create(g);
      // set miterlimit to 1 to avoid jaggy edges
      BasicStroke cs = (BasicStroke) graphics.getStroke();
      graphics.setStroke(new BasicStroke(cs.getLineWidth(), cs.getEndCap(), cs.getLineJoin(), 1.0f, cs.getDashArray(), cs.getDashPhase()));

      graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
              RenderingHints.VALUE_ANTIALIAS_ON);

      graphics.setColor(frameColor);
      graphics.setLineWidth(frameWidth);
      graphics.draw(shape);
      graphics.setLineWidth(lineWidth);
      graphics.setColor(color);
      graphics.draw(shape);
    }
  }


// -- Implementing HasPopupItems interface : -----------------------------------

  public JPopupMenu modifyPopupMenu(JPopupMenu menu, Component source, Point p) {
    Application.getApplication().getXMLMenuBuilder().mergePopupMenu( "wiredBasePopupMenu", menu);
    menu = graphicsPanel.modifyPopupMenu(menu, this);
    return menu;
  }

// -- Saving/Restoring : -------------------------------------------------------

  public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
    if (name != null) {
      nodeEl.setAttribute("name", name);
    }
    nodeEl.addContent(xmlioManager.save(graphicsPanel));
    if (handler != null) {
      nodeEl.addContent(xmlioManager.save(handler));
    }
    nodeEl.setAttribute("undo", undoSupport != null ? "true" : "false");
  }

  public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
    name = nodeEl.getAttributeValue("name");
    Iterator i = nodeEl.getChildren().iterator();
    setGraphicsPanel((GraphicsPanel) xmlioManager.restore((Element) i.next()));
    if (i.hasNext()) {
      setInteractionHandler((InteractionHandler) xmlioManager.restore((Element) i.next()));
    }
    if (nodeEl.getAttributeValue("undo", "").equals("true")) {
      addUndoSupport();
    }
  }
}
