// Copyright 2003-2006, FreeHEP.
package hep.wired.services;

import hep.wired.util.WiredRegistry;

import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;

import javax.swing.Icon;

import org.freehep.xml.io.XMLIO;

/**
 * Handles user interaction (Mouse and Keyboard) and dispatches edits to the
 * plot.
 * 
 * This interface filters the mouse interactions to only call subclasses for
 * button 1 mouse events. To keep cross platform compatibility it is wise to
 * only use button 1 mouse events (The Macintosh only has one mouse button).
 * Other buttons are mapped using special keys, but then again, keys can also be
 * used to adjust mouse behaviour. Also the pop up trigger (press and release)
 * is filtered and not acted on. The sequence of some mouse events is slightly
 * modified to make it easier to distinguish different sequences without extra
 * logic. The basic sequences are:
 * <UL>
 * <LI>press-release-click: mouseButton1Pressed(), mouseButton1Released(),
 * mouseButton1Clicked(); You would normally act in mouseButton1Clicked().
 * <LI>press-drag-release: mouseButton1Pressed(), mouseButton1DragStarted(),
 * mouseButton1Dragged(), mouseButton1DragEnded(), mouseButton1Released(). You
 * would normally act in mouseButton1DragStarted, Dragged and Ended.
 * mouseButton1DragStarted gets the same event as mouseButton1Pressed, but is
 * only called after the drag has started. It is immediately followed by a
 * mouseButton1Dragged(). mouseButton1DragEnded is called with the same event as
 * mouseButton1Released(), and is immediately followed by a call to the latter.
 * </UL>
 * All other sequences are as usual. Keyboard events are caught and handled. The
 * Esc key allows to abort when a mouse was already pressed, but not yet
 * released. The above sequences may be shortened to: press and press-drag. The
 * sublass' reset() method gets called for hitting the escape key. Key such as
 * ctrl, alt, shift and meta (option) are handled, and their repeat behaviour is filtered. Use
 * methods such as isAltKeyDown() to find out if the alt key is pressed.
 * 
 * @author Mark Donszelmann
 * @version $Id: InteractionHandler.java 8598 2006-08-14 20:38:02Z duns $
 */
public interface InteractionHandler extends WiredRegistry.ID, XMLIO {

    /**
     * Returns the name of the Interaction Handler.
     */
    public String getName();

    /**
     * Returns a one line description.
     */
    public String getDescription();
    
    /**
     * Returns true if this Interaction Handler is supported by this Graphics
     * Panel.
     */
    public boolean isSupportedBy(GraphicsPanel panel);

    /**
     * Sets the Record which is displayed in the GraphicsPanel. To be used if
     * the interaction handler is event dependent.
     */
    public void setRecord(RecordPlot plot, Object record);

    /**
     * Called when underlying plot is selected/unselected.
     */
    public void setSelected(RecordPlot plot, boolean selected);

    /**
     * Sets the size of the wired plot on the handler, to be able to move/resize
     * objects drawn by the handler.
     */
    public void setSize(RecordPlot plot, int width, int height);

    /**
     * Returns the width of the interaction panel.
     */
    public int getWidth();

    /**
     * Returns the height of the interaction panel.
     */
    public int getHeight();

    /**
     * Returns associated icon
     */
    public Icon getIcon(int size);

    /**
     * Called to reset the interaction handler in its initial state.
     */
    public void reset(RecordPlot plot, InputEvent event);

    /**
     * Called to allow the InteractionHandler to change the cursor.
     */
    public void changeCursor(RecordPlot plot, InputEvent event);

    /**
     * Called when mouse button 1 is clicked
     */
    public void mouseButton1Clicked(RecordPlot plot, MouseEvent event);

    /**
     * Called when mouse button 1 is pressed
     */
    public void mouseButton1Pressed(RecordPlot plot, MouseEvent event);

    /**
     * Called when mouse button 1 is released
     */
    public void mouseButton1Released(RecordPlot plot, MouseEvent event);

    /**
     * Called when mouse button 1 is starting a drag
     */
    public void mouseButton1DragStarted(RecordPlot plot, MouseEvent event);

    /**
     * Called when mouse button 1 is dragged
     */
    public void mouseButton1Dragged(RecordPlot plot, MouseEvent event);

    /**
     * Called when mouse button 1 is ending a drag
     */
    public void mouseButton1DragEnded(RecordPlot plot, MouseEvent event);

    /**
     * Called when the mouse is moved
     */
    public void mouseMoved(RecordPlot plot, MouseEvent event);

    /**
     * Called when the mouse is entering the component
     */
    public void mouseEntered(RecordPlot plot, MouseEvent event);

    /**
     * Called when the mouse is exitting the component
     */
    public void mouseExited(RecordPlot plot, MouseEvent event);

    /**
     * Called when the mousewheel is moved
     */
    public void mouseWheelMoved(RecordPlot plot, MouseWheelEvent event);

    /**
     * Called when the esc key is pressed. Return true if the event is handled.
     */
    public boolean escKeyPressed(RecordPlot plot, KeyEvent event);

    /**
     * Called when the esc key is released. Return true if the event is handled.
     */
    public boolean escKeyReleased(RecordPlot plot, KeyEvent event);

    /**
     * Called when the alt key is pressed. Return true if the event is handled.
     */
    public boolean altKeyPressed(RecordPlot plot, KeyEvent event);

    /**
     * Called when the alt key is released. Return true if the event is handled.
     */
    public boolean altKeyReleased(RecordPlot plot, KeyEvent event);

    /**
     * Called when the ctrl key is pressed. Return true if the event is handled.
     */
    public boolean ctrlKeyPressed(RecordPlot plot, KeyEvent event);

    /**
     * Called when the ctrl key is released. Return true if the event is
     * handled.
     */
    public boolean ctrlKeyReleased(RecordPlot plot, KeyEvent event);

    /**
     * Called when the shift key is pressed. Return true if the event is
     * handled.
     */
    public boolean shiftKeyPressed(RecordPlot plot, KeyEvent event);

    /**
     * Called when the esc key is released. Return true if the event is handled.
     */
    public boolean shiftKeyReleased(RecordPlot plot, KeyEvent event);

    /**
     * Called when the meta/option key is pressed. Return true if the event is
     * handled.
     * @since 4.1
     */
    public boolean metaKeyPressed(RecordPlot plot, KeyEvent event);

    /**
     * Called when the meta/option key is released. Return true if the event is handled.
     * @since 4.1
     */
    public boolean metaKeyReleased(RecordPlot plot, KeyEvent event);

    /**
     * Called when the ctrl (other OSs) or meta (MacOS) key is pressed key is pressed. Return true if the event is
     * handled. If handles ctrlKeyPressed or metaKeyPressed will NOT be called.
     * @since 4.1
     */
    public boolean ctrlMetaKeyPressed(RecordPlot plot, KeyEvent event);

    /**
     * Called when the ctrl (other OSs) or meta (MacOS) key is released. Return true if the event is handled.
     * If handled crlKeyPressed or metaKeyPressed will NOT be called.
     * @since 4.1 
     */
    public boolean ctrlMetaKeyReleased(RecordPlot plot, KeyEvent event);

    /**
     * Called when any other key is pressed. Return true if the event is
     * handled.
     */
    public boolean otherKeyPressed(RecordPlot plot, KeyEvent event);

    /**
     * Called when any other key is released. Return true if the event is
     * handled.
     */
    public boolean otherKeyReleased(RecordPlot plot, KeyEvent event);

    /**
     * Called when a key is typed.
     */
    public boolean keyTyped(RecordPlot plot, KeyEvent event);
}
