// Copyright 2005, FreeHEP.
package hep.wired.util.slider;


import hep.wired.util.SwingEventMulticaster;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Like DefaultBoundedRangeModel but for double values.
 * 
 * @author Mark Donszelmann
 * @version $Id: DefaultBoundedDoubleRangeModel.java 2139 2005-07-31 22:45:12Z
 *          duns $
 */
public class DefaultBoundedDoubleRangeModel implements BoundedDoubleRangeModel {

    private static final double epsilon = 0.00000000000001; 
    private boolean isAdjusting;
    private double minimum;
    private double maximum;
    private double def;
    private double value;
    private double extent;
    private ChangeListener listener;
    
    public DefaultBoundedDoubleRangeModel() {
        this(0, 0, 0, 100.0);
    }
    
    public DefaultBoundedDoubleRangeModel(double value, double extent, double min, double max) {
        if ((max >= min) && (value >= min) && ((value + extent) >= value)
                && ((value + extent) <= max)) {

            setRangeProperties(value, extent, min, max, value, false);
        } else {
            throw new IllegalArgumentException("invalid range properties");
        }
    }

    public boolean getValueIsAdjusting() {
        return isAdjusting; 
    }

    public void setValueIsAdjusting(boolean b) {
        setRangeProperties(getValue(), getExtent(), getMinimum(), getMaximum(), getDefaultValue(), b);
    }

    public void setMinimum(double min) {
        double newMax = Math.max(min, getMaximum());
        double newValue = Math.max(min, getValue());
        double newExtent = Math.min(newMax - newValue, getExtent());
        double newDef = Math.max(min, getDefaultValue());
        setRangeProperties(newValue, newExtent, min, newMax, newDef, getValueIsAdjusting());
    }

    public void setMaximum(double max) {
        double newMin = Math.min(max, getMinimum());
        double newExtent = Math.min(max - newMin, getExtent());
        double newValue = Math.min(max - newExtent, getValue());
        double newDef = Math.min(max, getDefaultValue());
        setRangeProperties(newValue, newExtent, newMin, max, newDef, getValueIsAdjusting());
    }

    public double getValue() {
        return value;
    }

    public void setValue(double value) {
        double newValue = Math.max(value, getMinimum());
        if (newValue + getExtent() > getMaximum()) {
            newValue = getMaximum() - getExtent();
        }
        setRangeProperties(newValue, getExtent(), getMinimum(), getMaximum(), getDefaultValue(), getValueIsAdjusting());
    }

    public double getExtent() {
        return extent;
    }

    public void setExtent(double extent) {
        double newExtent = Math.max(0, extent);
        if (getValue() + newExtent > getMaximum()) {
            newExtent = getMaximum() - getValue();
        }
        setRangeProperties(getValue(), newExtent, getMinimum(), getMaximum(), getDefaultValue(), getValueIsAdjusting());
    }

    public void setDefaultValue(double def) {
        setRangeProperties(getValue(), getExtent(), getMinimum(), getMaximum(), def, getValueIsAdjusting());
    }

    public void setRangeProperties(double newValue, double newExtent,
            double newMin, double newMax, double newDef, boolean newAdjusting) {
//         System.err.println("Range: "+newValue+" "+newExtent+" "+newMin+" "+newMax+" "+newAdjusting);

        if (newMin > newMax) {
            newMin = newMax;
        }
        if (newValue > newMax) {
            newMax = newValue;
        }
        if (newValue < newMin) {
            newMin = newValue;
        }

        if ((newExtent + newValue) > getMaximum()) {
            newExtent = getMaximum() - newValue;
        }
        if (newExtent < 0) {
            newExtent = 0;
        }

        boolean fire = DefaultBoundedDoubleRangeModel.rangePropertiesChanged(this, 
                               newValue, newExtent, newMin, newMax, newDef, newAdjusting);
        
        if (fire) {
            value = newValue;
            extent = newExtent;
            minimum = newMin;
            maximum = newMax;
            def = newDef;
            isAdjusting = newAdjusting;
            fireStateChanged();
        }
    }

    public static boolean rangePropertiesChanged(BoundedDoubleRangeModel model, double newValue, double newExtent,
            double newMin, double newMax, double newDef, boolean newAdjusting) {
        boolean deltaValue = Math.abs(newValue - model.getValue()) > epsilon;
        boolean deltaExtent = Math.abs(newExtent - model.getExtent()) > epsilon;
        boolean deltaMin = Math.abs(newMin - model.getMinimum()) > epsilon;
        boolean deltaMax = Math.abs(newMax - model.getMaximum()) > epsilon;
        boolean deltaDef = Math.abs(newDef - model.getDefaultValue()) > epsilon;
        boolean deltaAdjusting = newAdjusting != model.getValueIsAdjusting();
        return deltaValue || deltaExtent || deltaMin || deltaMax || deltaDef || deltaAdjusting;
    }
    
    public double getMinimum() {
        return minimum;
    }

    public double getMaximum() {
        return maximum;
    }

    public double getDefaultValue() {
        return def;
    }
    
    public void addChangeListener(ChangeListener a) {
        listener = SwingEventMulticaster.add(listener, a);
    }
    
    public void removeChangeListener(ChangeListener r) {
        listener = SwingEventMulticaster.remove(listener, r);
    }
    
    protected void fireStateChanged() {
        if (listener != null) {
            listener.stateChanged(new ChangeEvent(this));
        }
    }
}
