// Copyright 2005, FreeHEP.
package hep.wired.variable;

import java.lang.reflect.Field;

import org.freehep.xml.io.XMLIOManager;
import org.jdom.DataConversionException;
import org.jdom.Element;

/**
 * 
 * @author Mark Donszelmann
 * @version $Id: IntegerVariable.java 8598 2006-08-14 20:38:02Z duns $
 */

public class IntegerVariable extends NumberVariable {
    private int min, max, def;

    public IntegerVariable(String name, Object object, int min, int max,
            int def, String unit, String description) {
        super(name, object, unit, description);
        if (min > max)
            throw new IllegalArgumentException("Variable: min (" + min
                    + ") > max (" + max + ").");
        if ((def > max) || (def < min))
            throw new IllegalArgumentException("Variable: def (" + def
                    + ") not in interval [min,max] [" + min + "," + max + "].");
        this.min = min;
        this.max = max;
        this.def = def;
        setVariable(def);
    }

    /**
     * Returns the value of a named variable.
     */
    public int getIntegerVariable() {
        try {
            Field f = getObject().getClass().getDeclaredField(getName());
            f.setAccessible(true);
            if (f.getType() != int.class)
                throw new IllegalArgumentException("Variable with name '"
                        + getName() + "' not of type 'int'.");

            return f.getInt(getObject());
        } catch (NoSuchFieldException e) {
            throw new IllegalArgumentException("Variable with name '"
                    + getName() + "' does not exist as 'public' in class '"
                    + getObject().getClass() + "'.");
        } catch (IllegalAccessException e) {
            throw new IllegalArgumentException(
                    "No access to variable with name '" + getName() + "'.");
        }
    }

    public boolean setVariable(int value) {
        value = Math.max(value, getMinimum());
        value = Math.min(value, getMaximum());

        boolean fire = setValue(value);
        if (fire)
            fireStateChanged(VariableChangeEvent.VALUE);
        return fire;
    }

    private boolean setValue(int value) {
        try {
            Field f = getObject().getClass().getDeclaredField(getName());
            f.setAccessible(true);
            if (f.getType() == int.class) {
                if (f.getInt(getObject()) == value)
                    return false;
                f.setInt(getObject(), value);
            } else if (f.getType() == double.class) {
                if (f.getDouble(getObject()) == value)
                    return false;
                f.setDouble(getObject(), value);
            } else {
                throw new IllegalArgumentException("Variable with name '"
                        + getName() + "' not of type 'double' or 'int'.");
            }
        } catch (NoSuchFieldException e) {
            throw new IllegalArgumentException("Variable with name '"
                    + getName() + "' does not exist as 'public' in class '"
                    + getObject().getClass() + "'.");
        } catch (IllegalAccessException e) {
            throw new IllegalArgumentException(
                    "No access to variable with name '" + getName() + "'.");
        }
        return true;
    }

    public Object getValue() {
        return new Integer(getIntegerVariable());
    }

    public boolean setValue(Object value) {
        if (!(value instanceof Number)) {
            throw new IllegalArgumentException("Variable with name '"
                    + getName() + "' cannot handle value of class '"
                    + value.getClass() + "'.");
        }

        return setVariable(((Number) value).intValue());
    }

    public int getMinimum() {
        return min;
    }

    public void setMinimum(int min) {
        if (this.min == min)
            return;

        min = Math.min(min, max);
        this.min = min;
        int type = VariableChangeEvent.MIN;
        if (setValue(Math.max(min, getIntegerVariable())))
            type |= VariableChangeEvent.VALUE;
        fireStateChanged(type);
    }

    public int getMaximum() {
        return max;
    }

    public void setMaximum(int max) {
        if (this.max == max)
            return;

        max = Math.max(max, min);
        this.max = max;
        int type = VariableChangeEvent.MAX;
        if (setValue(Math.min(max, getIntegerVariable())))
            type |= VariableChangeEvent.VALUE;
        fireStateChanged(type);
    }

    public int getDefault() {
        return def;
    }

    public void saveValue(Element element) {
        element.setAttribute("val", Integer.toString(getIntegerVariable()));
    }

    public void restoreValue(Element element) {
        try {
            setVariable(element.getAttribute("val").getIntValue());
        } catch (DataConversionException dce) {
            throw new IllegalArgumentException(getClass() + ": "
                    + dce.getMessage());
        }
    }

    public String toString() {
        return Integer.toString(getIntegerVariable());
    }

    public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {

        nodeEl.setAttribute("min", Integer.toString(min));
        nodeEl.setAttribute("max", Integer.toString(max));
        nodeEl.setAttribute("def", Integer.toString(def));
    }

    public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        try {
            min = nodeEl.getAttribute("min").getIntValue();
            max = nodeEl.getAttribute("max").getIntValue();
            def = nodeEl.getAttribute("def").getIntValue();
        } catch (DataConversionException dce) {
            throw new IllegalArgumentException(getClass() + ": "
                    + dce.getMessage());
        }
    }
}
